#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "cgconfig.h"
#include "port.h"    
#include "shareinc.h"
#include "ifapi.h"
#include "fontapi.h"

#define CheckRET(a) { int x = a; if(x != 0) return x; }

#define offset_of(type, memb) ((int) ( (char *)&((type *)0)->memb - (char *)((type *)0) ))

#define MAX_CACHED_FONTS 10

/*  We can't figure out, when UFST uncaches its fonts.
    Also we need to cache font type outside UFST, because UFST
    takes it as input data. To resolve both problems
    we implement a simple cache, which restricts the opened font set
    with MAX_CACHED_FONTS.
*/

typedef struct font_cache_elem_s {
    char *font_path; 
    UW16 font_type;
    struct font_cache_elem_s * next;
    /*  fixme: this is designed for passing imbedded fonts through disk only. 
        Need a better font identification for passing through RAM.
    */
} font_cache_elem;

typedef struct {
    gs_plugin_interface ig;
} GXLIB_interface;

typedef struct renderer_instance_s {
    FAPI_interface If;
    GXLIB_interface Ig;
    int ref_count;
    int bInitialized;
    IF_STATE IFS;
    FONTCONTEXT fc;
    IFBITMAP * pbm;    
    double tran_xx, tran_xy, tran_yx, tran_yy;
    const char *decodingID;
    font_cache_elem * font_cache;
    FAPI_retcode (*construct)(struct renderer_instance_s *renderer);
    FAPI_retcode (*destruct)(struct renderer_instance_s *renderer);
} renderer_instance;

/* Type casts : */

static renderer_instance *Ig_to_I(gs_plugin_interface *Ig)
{   return (renderer_instance *)((char *)Ig - offset_of(renderer_instance, Ig));
}

static renderer_instance *If_to_I(gs_plugin_interface *If)
{   return (renderer_instance *)((char *)If - offset_of(renderer_instance, If));
}

static renderer_instance * to_instance(gs_plugin_interface *I0)
{   if (!strcmp(I0->interface_role, "GS"))
        return If_to_I(I0);
    else if (!strcmp(I0->interface_role, "GX"))
        return Ig_to_I(I0);
    assert(0);
    return 0;
}

/*----------------font-cache-members--------------------------------------*/

static int do_cache_font(renderer_instance *r, const char *font_path, UW16 font_type)
{   font_cache_elem *e = malloc(sizeof(font_cache_elem));
    if (e == NULL)
        return 0;
    e->font_path = strdup(font_path);
    e->font_type = font_type;
    e->next = r->font_cache;
    r->font_cache = e;
    return 1;
}

static void close_font(renderer_instance *r, const char *font_path)
{   FONTCONTEXT fc;
    UW16 code;
    memset(&fc, 0, sizeof(fc));
    fc.font_hdr = (char *)font_path;
    code = CGIFfont_purge(&r->IFS, &fc);
    (void)code;
    /* We ignore return code as a safe recover action. */
}

static int cache_font(renderer_instance *r, const char *font_path, UW16 font_type)
{   font_cache_elem **pe = &r->font_cache;
    if (*pe != NULL) {
        int count = 0;
        for (; (*pe)->next != NULL; pe = &(*pe)->next)
            count++;
        if (count >= MAX_CACHED_FONTS) {
            /* Uncache the last font : */
            close_font(r, (*pe)->font_path);
            free((*pe)->font_path);
            free(*pe);
            *pe = NULL;
        }
    }
    return do_cache_font(r, font_path, font_type);
}

static int get_cached_font_type(renderer_instance *r, const char *font_path, UW16 *font_type)
{   font_cache_elem *e = r->font_cache;
    for (; e != NULL; e = e->next)
        if (!strcmp(e->font_path, font_path)) {
            *font_type = e->font_type;
            return 1;
        }
    return 0;
}

static void font_cache_destroy(renderer_instance *r)
{   font_cache_elem *e = r->font_cache, *ee;
    r->font_cache = NULL;
    for (; e != NULL;) {
        ee = e->next;
        free(e->font_path);
        free(e);
        e = ee;
    }

}

/*------------------ renderer_instance members ------------------------------------*/

static FAPI_retcode construct(renderer_instance *r)
{   return 0; /* All done in make_instance. */
}		      

static FAPI_retcode destruct(renderer_instance *r)
{   if (r->bInitialized) {
        CGIFexit(&r->IFS);
        font_cache_destroy(r);
    }
    return 0;
}

/*------------------ FAPI_interface members ------------------------------------*/

static FAPI_retcode open(renderer_instance *r)
{   IFCONFIG   config_block;
    char *FAPI_USFT = getenv("FAPI_USFT");
    char *FAPI_USFT_TYPE = getenv("FAPI_USFT_TYPE");
    
    CheckRET(CGIFinit(&r->IFS));
    #if TT_ROM
        /* Not needed due to NO_SYMSET_MAPPING :
        LPUB8 ss_ptr[] = {NULL};
        CheckRET(CGIFinitRomInfo(&r->IFS, NULL, ss_ptr));
        */
    #endif
    strcpy(config_block.ufstPath, (FAPI_USFT != 0 ? FAPI_USFT : ""));
    strcpy(config_block.typePath, (FAPI_USFT_TYPE != 0 ? FAPI_USFT_TYPE : ""));
    config_block.num_files = 10;
    config_block.bit_map_width = 1;
    CheckRET(CGIFconfig(&r->IFS, &config_block));
    CheckRET(CGIFenter(&r->IFS));
    font_cache_destroy(r);
    return 0;
}		      

static FAPI_retcode ensure_open(gs_plugin_interface *fapi_interface)
{   renderer_instance *r = If_to_I(fapi_interface);
    if (r->bInitialized)
        return 0;
    r->bInitialized = 1;
    return open(r);
}		      

static UW16 get_font_type(const char *font_file_path)
{   char buf[20], mark_PS[]="%!";
    int i;
    FILE *f = fopen(font_file_path, "rb");
    if (f == NULL)
        return 0;
    if (fread(buf, 1, sizeof(buf), f) != sizeof(buf))
        return 0;
    fclose(f);
    for (i = 0; i < sizeof(buf) - sizeof(mark_PS); i++)
        if(!memcmp(buf + i, mark_PS, sizeof(mark_PS) - 1))
            return FC_PST1_TYPE;
    if (buf[0] == '\0' && buf[1] == '\1')
        return FC_TT_TYPE;
    if (buf[0] == 't' && buf[1] == 't')
        return FC_TT_TYPE;
    return 0;
}

static UW16 font_cached(renderer_instance *r, const char *font_file_path)
{   UW16 font_type;
    if (get_cached_font_type(r, font_file_path, &font_type))
        return font_type;
    font_type = get_font_type(font_file_path);
    if (!cache_font(r, font_file_path, font_type)) {
        /* This looks as memory had exhausted.
           We could do special processing here, 
           but UFST will return error code without our special assistance. 
           So recover it here with empty action. */
    }
    return font_type;
}
    
    
static int choose_decoding_PS(renderer_instance *r, FONTCONTEXT *fc, const char *cmapId)
{ /*    Work around UFST bug :
        It can't choose character by name due to bug with caching.
  */
  r->decodingID = "Latin1";
  return 1;
}

static int choose_decoding_TT(renderer_instance *r, FONTCONTEXT *fc, const char *cmapId)
{   int platId, specId, i;
    CMAP_QUERY q;
    if (sscanf(cmapId, "%d.%d", &platId, &specId) != 2)
        return 0;
    if (CGIFtt_cmap_query(&r->IFS, fc->font_hdr, fc->ttc_index, &q)) /* fixme : redundant calls. */
        return 0;
    for (i = 0; i < q.numCmap; i++)
        if (q.entry[i].platId == platId && q.entry[i].specId == specId) {
            fc->dl_ssnum = (specId << 4) | platId;
            return 1;
        }
    return 0;
}

static void scan_xlatmap(renderer_instance *r, FONTCONTEXT *fc, const char *xlatmap, const char *font_kind, 
                                    int (*choose_proc)(renderer_instance *r, FONTCONTEXT *fc, const char *cmapId))
{   const char *p = xlatmap;
    while(*p) {
        int good_kind =!strcmp(p, font_kind);
        p += strlen(p) + 2;
        while(*p) {
            const char *cmapId = p;
            r->decodingID = p + strlen(p) + 1;
            if (!*r->decodingID)
                break;
            p = r->decodingID + strlen(r->decodingID) + 1;
            if (good_kind)
                if (choose_proc(r, fc, cmapId))
                    return;
        }
    }
    r->decodingID = NULL;
}

static void choose_decoding(renderer_instance *r, FONTCONTEXT *fc, const char *xlatmap)
{   fc->dl_ssnum = fc->user_platID = fc->user_specID = 0;
    switch (fc->format & FC_FONTTYPE_MASK) {
        case FC_IF_TYPE: /* fixme */ break;
        case FC_PST1_TYPE: scan_xlatmap(r, fc, xlatmap, "PostScript", choose_decoding_PS); break;
        case FC_TT_TYPE: scan_xlatmap(r, fc, xlatmap, "TrueType", choose_decoding_TT); break;
        case FC_FCO_TYPE: /* fixme */ break;
    } 
}

static FAPI_retcode get_font(gs_plugin_interface *fapi_interface, const char *font_file_path, int subfont, const FracInt matrix[6], const FracInt HWResolution[2], const char *xlatmap, int bCache, void **font_handle)
{   renderer_instance *r = If_to_I(fapi_interface);
    FONTCONTEXT *fc = &r->fc;
    /*  Note : UFST doesn't provide handles for opened fonts,
        but copies FONTCONTEXT to IFSTATE and caches it.
        Due to this this plugin cannot provide a handle for the font.
        This assumes that only one font context is active at a moment.
    */
    /*  Fixme: bCache = 0 is not implemented yet.  */
    const double scale = F_ONE;
    double hx, hy, sx, sy;
    UW16 font_type = font_cached(r, font_file_path);
    r->decodingID = 0;
    r->tran_xx = matrix[0] / scale, r->tran_xy = matrix[1] / scale;
    r->tran_yx = matrix[2] / scale, r->tran_yy = matrix[3] / scale;
    hx = hypot(r->tran_xx, r->tran_xy), hy = hypot(r->tran_yx, r->tran_yy);
    sx = r->tran_xx * r->tran_yx + r->tran_xy * r->tran_yy; 
    sy = r->tran_xx * r->tran_yy - r->tran_xy * r->tran_yx;
    fc->font_id   = 0L;
    fc->xspot     = F_ONE;
    fc->yspot     = F_ONE;
    fc->fc_type   = FC_MAT2_TYPE;
    fc->s.m2.m[0] = (int)((double)matrix[0] / hx + 0.5);
    fc->s.m2.m[1] = (int)((double)matrix[1] / hx + 0.5);
    fc->s.m2.m[2] = (int)((double)matrix[2] / hy + 0.5);
    fc->s.m2.m[3] = (int)((double)matrix[3] / hy + 0.5);
    fc->s.m2.matrix_scale = 16;
    fc->s.m2.xworld_res = HWResolution[0];
    fc->s.m2.yworld_res = HWResolution[1];
    fc->s.m2.world_scale = 16;
    fc->s.m2.point_size   = (int)(hy * 8 + 0.5); /* 1/8ths of pixels */
    fc->s.m2.set_size     = (int)(hx * 8 + 0.5);
    fc->ssnum       = 0x8000; /* no symset mapping */
    //fc->ssnum       = RAW_GLYPH;
    fc->dl_ssnum    = 0;
    fc->user_platID = 0;
    fc->user_specID = 0;
    fc->font_hdr    = (LPUB8)font_file_path;
    fc->ttc_index   = subfont;
    fc->format      = font_type;
    choose_decoding(r, fc, xlatmap);
    fc->format      |= FC_NON_Z_WIND;   /* NON_ZERO Winding required for TrueType */
    fc->format      |= FC_INCHES_TYPE;  /* output in unitss per inch */
    //fc->format      |= FC_EXTERN_TYPE; /* passing through RAM */
    *font_handle = 0;
    return CGIFfont(&r->IFS, fc);
}

static FAPI_retcode release_font(gs_plugin_interface *fapi_interface, void *font_handle)
{   /*  This function must not be called, because bCache = 0 is not implemented. */
    return (FAPI_retcode)gs_plugin__rc_unimplemented;
}

static FAPI_retcode get_decodingID(gs_plugin_interface *fapi_interface, void *font_handle, const char **decodingID_result)
{   renderer_instance *r = If_to_I(fapi_interface);
    *decodingID_result = r->decodingID;
    return 0;
}

static FAPI_retcode get_font_bbox(gs_plugin_interface *fapi_interface, void *font_handle, int BBox[4])
{   renderer_instance *r = If_to_I(fapi_interface);
    SW16 VLCPower = 0;
    CheckRET(CGIFbound_box(&r->IFS, BBox, &VLCPower));
    BBox[0] >>= VLCPower;
    BBox[1] >>= VLCPower;
    BBox[2] >>= VLCPower;
    BBox[3] >>= VLCPower;
    return 0;
}

static FAPI_retcode can_retrive_char_by_name(gs_plugin_interface *fapi_interface, void *font_handle, int *result)
{   renderer_instance *r = If_to_I(fapi_interface);
    #if 0 /* UFST bug : it can retrive Type 1 by char name, but it needs char code as cache key. */
        *result = (r->fc.format & FC_FONTTYPE_MASK) == FC_PST1_TYPE;
    #else
        *result = 0;
    #endif
    return 0;
}

static FAPI_retcode get_char_width(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FracInt *wx, FracInt *wy)
{   renderer_instance *r = If_to_I(fapi_interface);
    UW16 buffer[2];
    UW16 code = CGIFwidth(&r->IFS, (UW16)c->char_code, 1, 4, buffer);
    *wx = (int)((double)buffer[0] * r->tran_xx / buffer[1] / 72.0 * r->fc.s.m2.xworld_res);
    *wy = (int)((double)buffer[0] * r->tran_xy / buffer[1] / 72.0 * r->fc.s.m2.yworld_res);
    return code;
}

static FAPI_retcode get_char_raster_metric(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_metrics *metrics)
{   renderer_instance *r = If_to_I(fapi_interface);
    IFBITMAP *pbm;
    int code;
    UW16 cc = (UW16)c->char_code;
    char PSchar_name[30];
    int len = min(sizeof(PSchar_name) - 1, c->char_name_length);
    memcpy(PSchar_name, c->char_name, len);
    PSchar_name[len] = 0;
    CheckRET(CGIFchIdptr(&r->IFS, &cc, PSchar_name));
    r->fc.format &= ~FC_OUTPUT_MASK; /* set bitmap format output */
    code = CGIFchar(&r->IFS, cc, &r->pbm, (SW16)0);
    pbm = r->pbm;
    if (code != ERR_fixed_space)
	CheckRET(code);
    metrics->esc_x = (int)((double)pbm->escapement * r->tran_xx / pbm->du_emx / 72.0 * r->fc.s.m2.xworld_res);
    metrics->esc_y = (int)((double)pbm->escapement * r->tran_xy / pbm->du_emy / 72.0 * r->fc.s.m2.yworld_res);
    if (code == ERR_fixed_space) {
	metrics->bbox_x0 = metrics->bbox_y0 = 0;
	metrics->bbox_x1 = metrics->bbox_y1 = -1;
	metrics->orig_x = metrics->orig_y = 0;
	metrics->lsb_x = metrics->lsb_y = 0;
	metrics->tsb_x = metrics->tsb_y = 0;
	return 0;
    }
    metrics->bbox_x0 = pbm->left_indent;
    metrics->bbox_y0 = pbm->top_indent;
    metrics->bbox_x1 = pbm->left_indent + pbm->black_width;
    metrics->bbox_y1 = pbm->top_indent + pbm->black_depth;
    metrics->orig_x = pbm->left_indent * 16 + pbm->xorigin;
    metrics->orig_y = pbm->top_indent  * 16 + pbm->yorigin;
    metrics->lsb_x = metrics->orig_x;
    metrics->lsb_y = 0;
    metrics->tsb_x = metrics->orig_x + pbm->black_width / 2;
    metrics->tsb_y = 0; /* pbm->topbearing; */
    return 0;
    /*	fixme : Current version of Agfa UFST cannot render enough metrics information
	without generating raster or outline. 
	If Agfa improves UFST API, this code will need to optimize.
    */
}

static int export_outline(renderer_instance *r, PIFOUTLINE pol, FAPI_path_interface *p)
{   POUTLINE_CHAR outchar;
    SW16 num_contrs,num_segmts;
    LPSB8 segment;
    PINTRVECTOR points;
    SW16  i,j;
    p->shift += r->If.frac_shift + pol->VLCpower;
    outchar = &pol->ol;
    num_contrs = outchar->num_loops;
    for(i=0; i<num_contrs; i++) {
     	num_segmts = outchar->loop[i].num_segmts;
        segment = (LPSB8)((LPSB8)(outchar->loop) + outchar->loop[i].segmt_offset);
        points = (PINTRVECTOR)((LPSB8)(outchar->loop) + outchar->loop[i].coord_offset);
        for(j=0; j<num_segmts; j++) {
            if(*segment == 0x00) {
             	CheckRET(p->moveto(p, points->x, points->y));
                points++;
            } else if (*segment == 0x01) {
		CheckRET(p->lineto(p, points->x, points->y));
                points++;
            } else if (*segment == 0x02) {
                points+=2;
                return -10; /* gs_error_invalidfont */
             	/* This must not happen */
            } else if (*segment == 0x03) {
		CheckRET(p->curveto(p, points[0].x, points[0].y, points[1].x, points[1].y, points[2].x, points[2].y));
                points+=3;
            } else {
                return -10; /* gs_error_invalidfont */
             	/* This must not happen */
	    }
            segment++;
        }
    }
    return 0;
}

static FAPI_retcode outline_char(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_path_interface *p, FAPI_metrics *metrics)
{   renderer_instance *r = If_to_I(fapi_interface);
    HIFOUTLINE hol= NIL_MH;
    PIFOUTLINE pol;
    r->fc.format &= ~FC_OUTPUT_MASK;
    r->fc.format |= FC_CUBIC_TYPE;       /* "cubic outline" output */
    CheckRET(CGIFchar_handle(&r->IFS, (UW16)c->char_code, (PHIFOUTLINE)&hol, (SW16)0)); /* fixme : char_name */
    pol = (PIFOUTLINE)MEMptr(hol);
    if (pol != 0) {
	CheckRET(export_outline(r, pol, p));
	metrics->bbox_x0 = pol->left;
	metrics->bbox_y0 = pol->top;
	metrics->bbox_x1 = pol->right;
	metrics->bbox_y1 = pol->bottom;
	metrics->orig_x = 0;
	metrics->orig_y = 0;
	metrics->lsb_x = 0;
	metrics->lsb_y = 0;
	metrics->tsb_x = (pol->left + pol->right) / 2;
	metrics->tsb_y = 0; /* pbm->topbearing; */
    } else
	memset(&metrics, 0, sizeof(metrics));
    if(hol)
        CHARfree(&r->IFS, hol);
    return 0;
}

#if 0 /* This is iseful if renderer doesn't cache chars. */
static int export_bitmap(LPUB8 t, SW16 ww, SW16 d, FAPI_raster *r)
{   int i,j;
    int h = (d < r->height ? d : r->height);
    int w = (ww < r->width ? ww << CHUNK_SHIFT : r->width);
    LPCHUNK tchnk = (LPCHUNK)t;
    int line_step =  (ww + (1 << (CHUNK_SHIFT - 3)) - 1) >> (CHUNK_SHIFT - 3);
    int num_chunks = (w  + (1 << CHUNK_SHIFT) - 1) >> CHUNK_SHIFT;
    for (i = 0; i < h; i++) {
	LPCHUNK o = (LPCHUNK)((char *)r->p + r->line_step * i);
	LPCHUNK s = tchnk + line_step * i;
        for (j=0; j < num_chunks; j++)
            *o++ = *s++;
    }
    return 0;
}
#endif

static FAPI_retcode get_char_raster(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_raster *rast)
{   renderer_instance *r = If_to_I(fapi_interface);
    IFBITMAP *pbm = r->pbm;
    if (pbm == 0)
	return 1003; /* Can't get raster */
    /*  CheckRET(export_bitmap((LPUB8)pbm->bm, pbm->width, pbm->depth, r));
	Use this code if renderer doesn't provide glyph cache.
    */
    rast->p = pbm->bm;
    rast->height = pbm->depth;
    rast->width = pbm->width << CHUNK_SHIFT;
    rast->line_step = (pbm->width + (1 << (CHUNK_SHIFT - 3)) - 1) >> (CHUNK_SHIFT - 3);
    return 0;
}

static FAPI_retcode release_char_raster(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c)
{   /* UFST has no function to release cached character. */ 
    renderer_instance *r = If_to_I(fapi_interface);
    r->pbm = 0;
    return 0;
}

static gs_plugin_interface *get_interface(gs_plugin_interface *I0, const char *interface_id, int minimal_version)
{   renderer_instance *I = to_instance(I0);
    if (!strcmp(interface_id, "GS")) {
        I->ref_count ++;
        return &I->If.ig;
    }
    if (!strcmp(interface_id, "GX")) {
        I->ref_count ++;
        return &I->Ig.ig;
    }
    return 0;
}

static gs_plugin_retcode make_instance(gs_plugin_factory *this, const char *interface_id, int minimal_major_version, gs_plugin_interface **p_instance);
static gs_plugin_retcode recycle_instance(gs_plugin_factory *f, renderer_instance *instance);
static gs_plugin_retcode can_close(gs_plugin_factory *f);
static gs_plugin_retcode close(gs_plugin_factory *f);

static gs_plugin_retcode release_interface(gs_plugin_interface *I0)
{   renderer_instance *I = to_instance(I0);
    if (I == 0)
        return gs_plugin__rc_typecheck;
    if(--I->ref_count == 0)
        recycle_instance(I0->factory, I);
    return gs_plugin__rc_version;
}

static gs_plugin_factory factory = {
    "FAPI",
    "AgfaUFST",
    "AgfaUFST-based font renderer for Ghostscript",
    1, 0, /* version */
    make_instance,
    can_close,
    close
};

static int instance_count = 0;

static gs_plugin_factory *factory_list[] = { &factory, 0 };

static FAPI_interface If0 = {
    {   &factory,
        "GS",
        get_interface,
        release_interface
    },
    16, /* frac_shift */
    1, /* renderer_caches_chars */
    ensure_open,
    get_font,
    release_font,
    get_decodingID,
    get_font_bbox,
    can_retrive_char_by_name,
    outline_char,
    get_char_width,
    get_char_raster_metric,
    get_char_raster,
    release_char_raster
};

static GXLIB_interface Ig0 = {
    {   &factory,
        "GX",
        get_interface,
        release_interface
    }
};

gs_plugin_factory ** get_factory_list()
{   return factory_list;
}

static gs_plugin_retcode make_instance(gs_plugin_factory *this, const char *interface_id, int minimal_major_version, gs_plugin_interface **p_instance)
{   renderer_instance *r = calloc(1, sizeof(renderer_instance));
    if (r == NULL)
        return gs_plugin__rc_memory;
    {   /*  We'd like to move this code to constructor, but we can't due to the order of definitions. 
            Actually we could, with few more efforts...
        */
        r->construct = construct;
        r->destruct = destruct;
        r->If = If0;
        r->Ig = Ig0;
    }
    CheckRET(r->construct(r));
    *p_instance = r->If.ig.get_interface(&r->If.ig, interface_id, 0);
    if (*p_instance == 0)
        return gs_plugin__rc_typecheck;
    instance_count ++;
    return gs_plugin__rc_success;
}		      

static gs_plugin_retcode recycle_instance(gs_plugin_factory *f, renderer_instance *r)
{   FAPI_retcode code;
    assert(f == &factory);
    code = r->destruct(r);
    free(r);
    instance_count --;
    return code;
}

static gs_plugin_retcode can_close(gs_plugin_factory *f)
{   assert(f == &factory);
    return (instance_count != 0 ? gs_plugin__rc_busy : gs_plugin__rc_success);
}

static gs_plugin_retcode close(gs_plugin_factory *f)
{   assert(can_close(f) == gs_plugin__rc_success);
    return gs_plugin__rc_success;
}
