#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "freetype.h"
#include "ifapi.h"
#include "fontapi.h"

#define CheckRET(a) { int x = a; if(x != 0) return x; }

#define offset_of(type, memb) ((int) ( (char *)&((type *)0)->memb - (char *)((type *)0) ))

#ifndef min
# define min(a,b)  ((a) < (b) ? (a) : (b))
#endif

typedef struct {
    gs_plugin_interface ig;
} GXLIB_interface;

typedef struct renderer_instance_s {
    FAPI_interface If;
    GXLIB_interface Ig;
    int ref_count;
    int bInitialized;
    FT_Library  lib;
    FAPI_retcode (*construct)(struct renderer_instance_s *renderer);
    FAPI_retcode (*destruct)(struct renderer_instance_s *renderer);
} renderer_instance;

/* Type casts : */

static renderer_instance *Ig_to_I(gs_plugin_interface *Ig)
{   return (renderer_instance *)((char *)Ig - offset_of(renderer_instance, Ig));
}

static renderer_instance *If_to_I(gs_plugin_interface *If)
{   return (renderer_instance *)((char *)If - offset_of(renderer_instance, If));
}

static renderer_instance * to_instance(gs_plugin_interface *I0)
{   if (!strcmp(I0->interface_role, "GS"))
        return If_to_I(I0);
    else if (!strcmp(I0->interface_role, "GX"))
        return Ig_to_I(I0);
    assert(0);
    return 0;
}

/*------------------ renderer_instance members ------------------------------------*/

static FAPI_retcode construct(renderer_instance *r)
{   return 0; /* All done in make_instance. */
}		      

static FAPI_retcode destruct(renderer_instance *r)
{   if (r->bInitialized)
        CheckRET(FT_Done_FreeType(r->lib));
    return 0;
}

/*------------------ the FAPI_interface members ------------------------------------*/

static FAPI_retcode ensure_open(gs_plugin_interface *fapi_interface)
{   renderer_instance *r = If_to_I(fapi_interface);
    if (r->bInitialized)
        return 0;
    r->bInitialized = 1;
    return FT_Init_FreeType(&r->lib);
}		      

static FAPI_retcode get_font(gs_plugin_interface *fapi_interface, const char *font_file_path, int subfont, const FracInt matrix[6], const FracInt HWResolution[2], const char *xlatmap, int bCache, void **font_handle)
{   renderer_instance *r = If_to_I(fapi_interface);
    FT_Library  lib = r->lib;
    FT_Matrix m; /* 16.16 */
    FT_Vector v; /* 26.6 */
    const double s6 = 1 << 6, s16 = 1 << 16;
    const double scale = s6;
    double hx, hy, sx, sy;
    double tran_xx, tran_xy, tran_yx, tran_yy;
    FT_Fixed hwrx = (FT_Fixed)(HWResolution[0]/scale + 0.5);
    FT_Fixed hwry = (FT_Fixed)(HWResolution[1]/scale + 0.5);
    FT_Fixed w, h;
    tran_xx = matrix[0] / scale, tran_xy = matrix[1] / scale;
    tran_yx = matrix[2] / scale, tran_yy = matrix[3] / scale;
    hx = hypot(tran_xx, tran_xy), hy = hypot(tran_yx, tran_yy);
    sx = tran_xx * tran_yx + tran_xy * tran_yy; 
    sy = tran_xx * tran_yy - tran_xy * tran_yx;
    w = (FT_Fixed)(hx * s6 + 0.5);
    h = (FT_Fixed)(fabs(sy / hx * s6) + 0.5);
    m.xx = (FT_Fixed)(tran_xx / w * s6 * s16 + 0.5);
    m.yx = (FT_Fixed)(tran_xy / w * s6 * s16 + 0.5); /* Note: yx <- xy */
    m.xy = (FT_Fixed)(tran_yx / h * s6 * s16 + 0.5);
    m.yy = (FT_Fixed)(tran_yy / h * s6 * s16 + 0.5);
    v.x = (FT_Fixed)(matrix[4] / (s16 / s6) + 0.5);
    v.y = (FT_Fixed)(matrix[5] / (s16 / s6) + 0.5);
    CheckRET(FT_New_Face(lib, font_file_path, subfont, (FT_Face *)font_handle));
    CheckRET(FT_Set_Char_Size(*(FT_Face *)font_handle, w, h, hwrx, hwry));
    FT_Set_Transform(*(FT_Face *)font_handle, &m, &v);
    return 0;
}

static FAPI_retcode release_font(gs_plugin_interface *fapi_interface, void *font_handle)
{   /*  This function must not be called, because bCache = 0 is not implemented. */
    return gs_plugin__rc_unimplemented;
}

static FAPI_retcode get_decodingID(gs_plugin_interface *fapi_interface, void *font_handle, const char **decodingID_result)
{   return 0;
}

static FAPI_retcode get_font_bbox(gs_plugin_interface *fapi_interface, void *font_handle, int BBox[4])
{   FT_Face f = (FT_Face)font_handle;
    BBox[0] = (int)((double)f->bbox.xMin / f->units_per_EM * 1000 - 0.5);
    BBox[1] = (int)((double)f->bbox.yMin / f->units_per_EM * 1000 - 0.5);
    BBox[2] = (int)((double)f->bbox.xMax / f->units_per_EM * 1000 + 0.5);
    BBox[3] = (int)((double)f->bbox.yMax / f->units_per_EM * 1000 + 0.5);
    return 0;
}

static FAPI_retcode can_retrive_char_by_name(gs_plugin_interface *fapi_interface, void *font_handle, int *result)
{   
    return 0;
}

static FAPI_retcode get_char_width(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FracInt *wx, FracInt *wy)
{   return 0;
}

static FAPI_retcode get_char_raster_metric(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_metrics *metrics)
{   FT_Face f = (FT_Face)font_handle;
    FT_Glyph_Metrics *m;
    FT_GlyphSlot s;

    FT_ULong cc = (FT_ULong)c->char_code;
    char PSchar_name[30]; /* fixme : FreeType has no function for this. */
    int len = min(sizeof(PSchar_name) - 1, c->char_name_length);
    memcpy(PSchar_name, c->char_name, len);
    PSchar_name[len] = 0;

    CheckRET(FT_Load_Char(f, cc, FT_LOAD_DEFAULT));
    s = f->glyph;
    m = &s->metrics;
    metrics->bbox_x0 = m->horiBearingX;
    metrics->bbox_y0 = m->height - m->horiBearingY;
    metrics->bbox_x1 = metrics->bbox_x0 + m->width;
    metrics->bbox_y1 = metrics->bbox_y0 + m->height;
    metrics->orig_x = s->bitmap_left * 16 + m->horiBearingX / 4;
    metrics->orig_y = s->bitmap_top * 16 + (m->height - m->horiBearingY) / 4;
    metrics->lsb_x = m->horiBearingX;
    metrics->lsb_y = m->height - m->horiBearingY;
    metrics->tsb_x = m->vertBearingX; /* ? */
    metrics->tsb_y = m->vertBearingY; /* ? */
    metrics->esc_x = f->glyph->advance.x;
    metrics->esc_y = f->glyph->advance.y;
    /* vertAdvance ?? */
    return 0;
}

static FAPI_retcode outline_char(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_path_interface *p, FAPI_metrics *metrics)
{   return gs_plugin__rc_unimplemented; /* fixme */
} 

static FAPI_retcode get_char_raster(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c, FAPI_raster *rast)
{   FT_Face f = (FT_Face)font_handle;
    FT_GlyphSlot s = f->glyph;
    FT_Bitmap *b = &s->bitmap;
    CheckRET(FT_Render_Glyph(s, ft_render_mode_mono));
    rast->p = b->buffer;
    rast->height = b->rows;
    rast->width = b->width;
    rast->line_step = b->pitch;
    return 0;
}

static FAPI_retcode release_char_raster(gs_plugin_interface *fapi_interface, void *font_handle, FAPI_char_ref *c)
{   /* FreeType2 has no function for releaseing cached character. */ 
    return 0;
}

static gs_plugin_interface *get_interface(gs_plugin_interface *I0, const char *interface_id, int minimal_version)
{   renderer_instance *I = to_instance(I0);
    if (!strcmp(interface_id, "GS")) {
        I->ref_count ++;
        return &I->If.ig;
    }
    if (!strcmp(interface_id, "GX")) {
        I->ref_count ++;
        return &I->Ig.ig;
    }
    return 0;
}

static gs_plugin_retcode make_instance(gs_plugin_factory *this, const char *interface_id, int minimal_major_version, gs_plugin_interface **p_instance);
static gs_plugin_retcode recycle_instance(gs_plugin_factory *f, renderer_instance *instance);
static gs_plugin_retcode can_close(gs_plugin_factory *f);
static gs_plugin_retcode close(gs_plugin_factory *f);

static gs_plugin_retcode release_interface(gs_plugin_interface *I0)
{   renderer_instance *I = to_instance(I0);
    if (I == 0)
        return gs_plugin__rc_typecheck;
    if(--I->ref_count == 0)
        recycle_instance(I0->factory, I);
    return gs_plugin__rc_success;
}

static gs_plugin_factory factory = {
    "FAPI",
    "FreeType2",
    "FreeType2-based font renderer for Ghostscript",
    1, 0, /* version */
    make_instance,
    can_close,
    close
};

static int instance_count = 0;

static gs_plugin_factory *factory_list[] = { &factory, 0 };

static FAPI_interface If0 = {
    {   &factory,
        "GS",
        get_interface,
        release_interface
    },
    6, /* frac_shift */
    1, /* renderer_caches_chars */
    ensure_open,
    get_font,
    release_font,
    get_decodingID,
    get_font_bbox,
    can_retrive_char_by_name,
    outline_char,
    get_char_width,
    get_char_raster_metric,
    get_char_raster,
    release_char_raster
};

static GXLIB_interface Ig0 = {
    {   &factory,
        "GX",
        get_interface,
        release_interface
    }
};

gs_plugin_factory ** get_factory_list()
{   return factory_list;
}

static gs_plugin_retcode make_instance(gs_plugin_factory *this, const char *interface_id, int minimal_major_version, gs_plugin_interface **p_instance)
{   renderer_instance *r = calloc(1, sizeof(renderer_instance));
    *p_instance = (gs_plugin_interface *)r;
    if (r == NULL)
        return (FAPI_retcode)gs_plugin__rc_memory;
    {   /*  We'd like to move this code to constructor, but we can't due to the order of definitions. 
            Actually we could with few more efforts...
        */
        r->construct = construct;
        r->destruct = destruct;
        r->If = If0;
        r->Ig = Ig0;
    }
    CheckRET(r->construct(r));
    *p_instance = r->If.ig.get_interface(&r->If.ig, interface_id, 0);
    if (*p_instance == 0)
        return gs_plugin__rc_typecheck;
    instance_count ++;
    return gs_plugin__rc_success;
}		      

static gs_plugin_retcode recycle_instance(gs_plugin_factory *f, renderer_instance *r)
{   FAPI_retcode code;
    assert(f == &factory);
    code = r->destruct(r);
    free(r);
    instance_count --;
    return code;
}

static gs_plugin_retcode can_close(gs_plugin_factory *f)
{   assert(f == &factory);
    return (instance_count != 0 ? gs_plugin__rc_busy : gs_plugin__rc_success);
}

static gs_plugin_retcode close(gs_plugin_factory *f)
{   assert(can_close(f) == gs_plugin__rc_success);
    return gs_plugin__rc_success;
}



