/* Copyright (C) 1989, 1995, 1996, 1997, 1998, 1999 Aladdin Enterprises.  All rights reserved.
  
  This file is part of AFPL Ghostscript.
  
  AFPL Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author or
  distributor accepts any responsibility for the consequences of using it, or
  for whether it serves any particular purpose or works at all, unless he or
  she says so in writing.  Refer to the Aladdin Free Public License (the
  "License") for full details.
  
  Every copy of AFPL Ghostscript must include a copy of the License, normally
  in a plain ASCII text file named PUBLIC.  The License grants you the right
  to copy, modify and redistribute AFPL Ghostscript, but only under certain
  conditions described in the License.  Among other things, the License
  requires that the copyright notice and this notice be preserved on all
  copies.
*/

/*$Id: gsplugin.h,v 1.0 2000/09/19 19:00:32 lpd Exp $ */
/* generic plugin interface */

#ifndef gsplugin_INCLUDED
#  define gsplugin_INCLUDED

#include "gsdllapi.h"

/*  GS plugin interface is an artifact being similar to
    Component Object Model (COM). Rather it provides similar
    functionality as COM, it differs from COM with
    some generalization and simplification.
*/

/* Some predefined types (see comments below) : */
typedef struct gs_plugin_factory_s gs_plugin_factory;
typedef struct gs_plugin_interface_s gs_plugin_interface;
typedef gs_plugin_factory ** (* gs_get_factories)(P0());

/* Return codes for plugins : */
typedef enum {
    gs_plugin__rc_success       = 0,
    gs_plugin__rc_excess        = -1001,
    gs_plugin__rc_version       = -1002,
    gs_plugin__rc_memory        = -1003,
    gs_plugin__rc_busy          = -1004,
    gs_plugin__rc_unimplemented = -1005,
    gs_plugin__rc_typecheck     = -1006,
    gs_plugin__rc_export        = -1007
} gs_plugin_retcode;

/*  Plugin factory is plugin's static data.
    It makes and recycles plugin instances.
    Factory to be derived form gs_plugin_factory.
    Also it provides runtime information about the type of 
    plugin and its version.

    Plugin interface to be derived from gs_plugin_interface.
    Plugin instance is allowed to have multiple interfaces for
    different purposes. For example, plugin instance
    may have 3 interfaces for communicating with 
    Graphics Library, with Ghostscript, and with PCL interpreter,
    rather it may happen that all 3 interfaces are same.

    The tririplet (factory->type, factory->version_high, interface_type) 
    must identify the instance class as C languadge type. In other words,
    this tririplet is runtime type information for interface pointer.

    Note that we don't put interface_type to gs_plugin_interface.
    We put interface_role instead. This allows to use multiple
    inheritance and virtual bases subcalssing interfaces.
    Interface_role is runtime information about the role of
    interface in the instance, being used for type casts.

    We don't define base class for plugin instance,
    because no generic functionality is exposed for clients.
    Moreover, plugin instanse itself is not visible for clients.
    Clients access it through interfaces only.
    There is no assumptions, what plugin instance does, how it behaves, 
    and what data it contains. 

    Plugin instance should implement internal counter,
    which increases each time when get_interface is called
    for any of the instance's interfaces, and decreases when
    release_interface is called for any of instance's interfaces.
    When the counter drops to zero, the instance to be 
    recycled with factory. It is internal plugin's deal,
    how to recycle instance.
    
    Factory's outlet "make_instance" may create instances dynamically 
    or obtain them from a precreated pool, or else. Factory's recycler 
    (being invisible for clients) may completely destroy instance 
    and release its resources, or return it to the pool 
    for future use. It's internal plugin's deal, how to provide 
    the consistency between outlet and recicler, but the consistency
    is obligatory.

    Factory should open automatically on first call to make_instance.
    It can close when all instances are recycled. To provide consistent closing,
    factory should implement instance counting internally.

    Regular plugin DLL may define several factories 
    of various types, and must export function "get_factories", which
    retrives zero-ended array of pointers to factories. 
    Clients (such as Ghostscript or Graphics Library)
    should choose factories from the array with associative search,
    using factory->type (and possibly factory->subtype) as key,
    and the check obtained version. 
    Multiple factories are allowed to share code between plugins 
    of different types, being linked statically within a DLL.

    In this abstraction layer we don't specify functionality 
    for managing plugin DLLs. The reason is that DLL cannot unload itself,
    therefore plugin DLL manager to be coded in a separate DLL.
    Our temporary solution is to put the manager into Graphics Library -
    see gsplugin.c .
*/

struct gs_plugin_factory_s {
    const char *type;          /* Plugin type, such as FontAPI - describes purpose of instances. */
    const char *subtype;       /* Plugin subtype, such as AgfaUFST - describes implementation of instances. */
    const char *description;   /* Plugin description in natural language - used in messages. */
    const short version_major; /* Plugin major version - identifies plugin interface type, to be checked before calls to plugin. */
    const short version_minor; /* Plugin minor version - used in messages. */
    gs_plugin_retcode (GSDLLAPIPTR make_instance)(P4(gs_plugin_factory *this, const char *interface_type, int minimal_major_version, gs_plugin_interface **p_interface)); /* outlet */
    gs_plugin_retcode (GSDLLAPIPTR can_close)(P1(gs_plugin_factory *this));
    gs_plugin_retcode (GSDLLAPIPTR close)(P1(gs_plugin_factory *this));
};

struct gs_plugin_interface_s {
    gs_plugin_factory *factory;
    char *interface_role;
    gs_plugin_interface * (GSDLLAPIPTR get_interface)(P3(gs_plugin_interface *this, const char *interface_type, int minimal_major_version));
    gs_plugin_retcode (GSDLLAPIPTR release_interface)(P1(gs_plugin_interface *this));
};

typedef gs_plugin_factory ** (GSDLLAPIPTR PFN_get_factories)(P1(int version_major));

#endif
