/* Copyright (C) 1991, 2000 Aladdin Enterprises.  All rights reserved.
  
  This file is part of AFPL Ghostscript.
  
  AFPL Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author or
  distributor accepts any responsibility for the consequences of using it, or
  for whether it serves any particular purpose or works at all, unless he or
  she says so in writing.  Refer to the Aladdin Free Public License (the
  "License") for full details.
  
  Every copy of AFPL Ghostscript must include a copy of the License, normally
  in a plain ASCII text file named PUBLIC.  The License grants you the right
  to copy, modify and redistribute AFPL Ghostscript, but only under certain
  conditions described in the License.  Among other things, the License
  requires that the copyright notice and this notice be preserved on all
  copies.
*/

/*$Id: zfontapi.c,v 1.0 2000/12/03 23:35:30 lpd Exp $ */
/* Font API client */

#include <stdlib.h>
#include "math_.h"
#include "ghost.h"
#include "gp.h"
#include "oper.h"
#include "gxdevice.h"
#include "gxfont.h"
#include "gxchar.h"
#include "bfont.h"
#include "dstack.h"
#include "ichar.h"
#include "idict.h"
#include "iddict.h"
#include "iname.h"
#include "igstate.h"
#include "ifapi.h"
#include "iplugins.h"
#include "store.h"
#include "gzstate.h"
#include "stream.h"		/* for files.h */
#include "files.h"

private int bCacheFonts = 1; /* GS cannot provide release_font signal. */
 
#define CheckRET(a) { int x = a; if(x < 0) return x; }

static struct FAPI_outline_handler_s {
    struct gx_path_s *path;
    fixed x0, y0;
} olh;

private inline int import_shift(int x, int n)
{   return n > 0 ? x << n : x >> -n;
}

private int add_move(FAPI_path_interface *I, FracInt x, FracInt y)
{   return gx_path_add_point(I->olh->path, import_shift(x,I->shift) + I->olh->x0, -import_shift(y,I->shift) + I->olh->y0);
}

private int add_line(FAPI_path_interface *I, FracInt x, FracInt y)
{   return gx_path_add_line_notes(I->olh->path, import_shift(x,I->shift) + I->olh->x0, -import_shift(y,I->shift) + I->olh->y0, 0);
}

private int add_curve(FAPI_path_interface *I, FracInt x0, FracInt y0, FracInt x1, FracInt y1, FracInt x2, FracInt y2)
{   return gx_path_add_curve_notes(I->olh->path, import_shift(x0,I->shift) + I->olh->x0, -import_shift(y0,I->shift) + I->olh->y0, 
					         import_shift(x1,I->shift) + I->olh->x0, -import_shift(y1,I->shift) + I->olh->y0, 
					         import_shift(x2,I->shift) + I->olh->x0, -import_shift(y2,I->shift) + I->olh->y0, 0);
}

private int add_closepath(FAPI_path_interface *I)
{   return gx_path_close_subpath_notes(I->olh->path, 0);
}

private FAPI_path_interface path_interface = { &olh, 0, add_move, add_line, add_curve, add_closepath };

private int FAPI_get_xlatmap(i_ctx_t *i_ctx_p, char **xlatmap)
{   ref *pref;
    CheckRET(dict_find_string(systemdict, ".xlatmap", &pref));
    if (r_type(pref) != t_string)
        return_error(e_typecheck);
    *xlatmap = (char *)pref->value.bytes;
    /*  Note : this supposes that xlatmap doesn't move in virtual memory.
        Garbager must not be called while plugin executes get_font, get_decodingID.
        To be fixed with making copy of xlatmap in system memory.
    */
    return 0;
}

private int renderer_retcode(i_ctx_t *i_ctx_p, FAPI_interface *I, FAPI_retcode rc)
{   stream *s;
    if (rc == 0)
	return 0;
    if (zget_stdout(i_ctx_p, &s) == 0) {
	uint wlen;
	char *s0 = "Error: Font Renderer Plugin ( ", *s1 = " ) return code = ", buf[10];
	sprintf(buf, "%d", rc);
	sputs(s, (byte *)s0, strlen(s0), &wlen);
	sputs(s, (byte *)I->ig.factory->subtype, strlen(I->ig.factory->subtype), &wlen);
	sputs(s, (byte *)s1, strlen(s1), &wlen);
	sputs(s, (byte *)buf, strlen(buf), &wlen);
	sputs(s, (byte *)"\n", 1, &wlen);
    }
    return_error(e_invalidfont);
}

private int FAPI_find_plugin(i_ctx_t *i_ctx_p, const char *subtype, FAPI_interface **pI)
{   int i;
    gs_plugin_interface **plugins = i_get_installed_plugins(i_ctx_p);
    for (i = 0; plugins[i] != 0; i++)
       	if (!strcmp(plugins[i]->factory->type,"FAPI"))
	    if (!strcmp(plugins[i]->factory->subtype, subtype)) {
	        FAPI_interface *I = *pI = (FAPI_interface *)plugins[i];
                if (I->ig.factory->version_major < 1)
                    return renderer_retcode(i_ctx_p, I, (int)gs_plugin__rc_version);
	        CheckRET(renderer_retcode(i_ctx_p, I, I->ensure_open(&I->ig)));
	        return 0;
	    }
    return_error(e_invalidfont);
} 

private int FAPI_refine_font(i_ctx_t *i_ctx_p, os_ptr op, gs_font_base *pbfont, char *font_file_path)
{   ref *pdr = osp;  /* font dict */
    dict *pdict = pdr->value.pdict;
    FracInt matrix[6] = {1, 0, 0, 1, 0, 0};
    FAPI_interface *I;
    void *API_font_handle;
    double size, size1;
    ref *v, x0, y0, x1, y1, Decoding, *SubfontId;
    int HWResolution[2], BBox[4], scale, subfont = 0;
    char *decodingID = NULL;
    char *xlatmap; 
    CheckRET(FAPI_get_xlatmap(i_ctx_p, &xlatmap));
    CheckRET(FAPI_find_plugin(i_ctx_p, pbfont->FAPI_ID, &I));
    scale = 1 << I->frac_shift;
    size1 = size = 1 / hypot(pbfont->FontMatrix.xx, pbfont->FontMatrix.xy);
    if (size < 1000)
	size = 1000;
    if (size1 > 100)
        size1 = (int)(size1 + 0.5);
    matrix[0] = matrix[3] = (int)(size * scale + 0.5);
    HWResolution[0] = (FracInt)(72 * scale);
    HWResolution[1] = (FracInt)(72 * scale);
    if (dict_find_string(pdr, "SubfontId", &SubfontId) >= 0 && r_has_type(SubfontId, t_integer))
        subfont = SubfontId->value.intval;
    CheckRET(renderer_retcode(i_ctx_p, I, (I->get_font)(&I->ig, font_file_path, subfont, matrix, HWResolution, xlatmap, bCacheFonts, &API_font_handle)));
    CheckRET(renderer_retcode(i_ctx_p, I, (I->get_font_bbox)(&I->ig, &API_font_handle, BBox)));
    CheckRET(renderer_retcode(i_ctx_p, I, (I->get_decodingID)(&I->ig, &API_font_handle, &decodingID)));

    /* Refine FontBBox : */
    pbfont->FontBBox.p.x = (double)BBox[0] * size1 / size;
    pbfont->FontBBox.p.y = (double)BBox[1] * size1 / size;
    pbfont->FontBBox.q.x = (double)BBox[2] * size1 / size;
    pbfont->FontBBox.q.y = (double)BBox[3] * size1 / size;
    if (dict_find_string(op, "FontBBox", &v) < 0 || !r_has_type(v, t_array))
	return_error(e_invalidfont);
    if (r_size(v) < 4)
	return_error(e_invalidfont);
    make_real(&x0, pbfont->FontBBox.p.x);
    make_real(&y0, pbfont->FontBBox.p.y);
    make_real(&x1, pbfont->FontBBox.q.x);
    make_real(&y1, pbfont->FontBBox.q.y);
    ref_assign_old(v, v->value.refs + 0, &x0, "FAPI_refine_font_BBox");
    ref_assign_old(v, v->value.refs + 1, &y0, "FAPI_refine_font_BBox");
    ref_assign_old(v, v->value.refs + 2, &x1, "FAPI_refine_font_BBox");
    ref_assign_old(v, v->value.refs + 3, &y1, "FAPI_refine_font_BBox");

    /* Assign Decoding : */
    if (decodingID != 0 && *decodingID) {
        CheckRET(name_ref((byte *)decodingID, strlen(decodingID), &Decoding, 0));
        CheckRET(dict_put_string(pdr, "Decoding", &Decoding, NULL));    /* Insert entry to font dictionary. */
    }
    return 0;
}

int notify_remove_file(void *proc_data, void *event_data)
{   /* gs_font_finalize passes event_data == NULL, so check it here. */
    if (event_data == NULL) {
        char *path = (char *)proc_data;
        remove(path);
    }
    return 0;
}

/* <string|name> <font_dict> .buildfontFAPI <string|name> <font> */
/* Build a font being handled by external font API. */
private int zbuildfontFAPI(i_ctx_t *i_ctx_p)
{   os_ptr op = osp;
    build_proc_refs build;
    gs_font_base *pfont;
    font_type ftype;
    ref *v;
    char *font_file_path;
    const byte *pchars;
    int code;
    uint len;

    CheckRET(build_proc_name_refs(&build, ".FAPIBuildChar", ".FAPIBuildGlyph"));
    CheckRET(dict_int_param(op, "FontType", 0, 42, 0, &ftype));
    CheckRET(build_gs_outline_font(i_ctx_p, op, &pfont, ftype, &st_gs_font_base, &build, 0, build_gs_simple_font));

    if (dict_find_string(op, "FAPI", &v) < 0 || !r_has_type(v, t_name))
	return_error(e_invalidfont);
    obj_string_data(v, &pchars, &len);
    len = min(len, sizeof(pfont->FAPI_ID) -1);
    strncpy(pfont->FAPI_ID, (char *)pchars, len);
    pfont->FAPI_ID[len] = 0;

    if (dict_find_string(op, "Path", &v) < 0 || !r_has_type(v, t_string))
	return_error(e_invalidfont);
    font_file_path = ref_to_string(v, imemory_global, "font file path");
    code = FAPI_refine_font(i_ctx_p, op, pfont, font_file_path);
    if (dict_find_string(op, "Temp", &v) <= 0)
        gs_free_string(imemory_global, (byte *)font_file_path, r_size(v) + 1, "font file path");
    else {   
        int code1 = gs_notify_register(&pfont->notify_list, notify_remove_file, font_file_path);
        (void)code1; /* No chance to recover */
    }
    CheckRET(code);
    return define_gs_font((gs_font *)pfont);
}

private void remove_all_files(char *path_wild, int pl, gs_memory_t * mem)
{   int maxname = 256;  
    int L = (pl + maxname) * 2, code;
    char *buf;
    file_enum *pfen = gp_enumerate_files_init(path_wild, strlen(path_wild), mem);
    if (pfen == NULL)
        return; /* No chance to repair or report error. */
    buf = (char *)gs_alloc_string(mem, L, "temp font path");
    if (buf != NULL) {
        strcpy(buf, path_wild);
        while (code = gp_enumerate_files_next(pfen, buf, L), code > 0 && code != ~(uint)0) {
            //remove(buf);
        }
        gs_free_string(mem, (byte *)buf, L, "temp font path");
    }
    gp_enumerate_files_close(pfen);
}


private int FAPI_do_char(i_ctx_t *i_ctx_p, gx_device *dev, char *font_file_path)
{   /* Stack : <font> <code|name> --> - */
    os_ptr op = osp;
    gs_font *pfont;
    int code = font_param(op - 1, &pfont);
    const gs_font_base *pbfont = (gs_font_base *) pfont;
    gs_text_enum_t *penum = op_show_find(i_ctx_p);
    gs_show_enum *penum_s = (gs_show_enum *)penum; /* fixme : is this ever correct ? */
    /* 
        fixme: the following code needs to optimize with maintaining scaled font objects 
        in graphics library and in interpreter. Now we suppose that the renderer
        uses font cache, so as redundant font opening isn't big expense.
    */
    FAPI_char_ref cr = {0, 0, 0};
    const gs_matrix * ctm = &ctm_only(igs), *base_font_matrix = &pfont->base->FontMatrix;
    int scale, unscale;
    int HWResolution[2];
    FracInt matrix[6]; /* Font to user space transformation */
    FAPI_metrics metrics;
    FAPI_interface *I;
    void *API_font_handle;
    int client_char_code;
    bool char_by_name = true;
    ref char_name, *SubfontId;
    char *xlatmap; 
    int subfont = 0;
    CheckRET(FAPI_get_xlatmap(i_ctx_p, &xlatmap)); /* Fixme : redundant calls. */
    if (penum == 0)
	return_error(e_undefined);
    if (r_has_type(op, t_integer)) {
	/* Translate from PS encoding to char name : */
        /* fixme: this is only useful for Type 1. */
	ref *Encoding;
	int_param(op, 0xFF, &client_char_code);
        if (dict_find_string(osp - 1, "Encoding", &Encoding) >= 0 && r_has_type(Encoding, t_array)) {
	    if (array_get(Encoding, client_char_code, &char_name) < 0)
	        CheckRET(name_ref((byte *)".notdef", 7, &char_name, -1));
	} else {
	    /* The font has no Encoding, go with client's char code without translation : */
	    char_by_name = false;
	    cr.char_code = client_char_code;
	}
    } else 
	char_name = *op;
    CheckRET(FAPI_find_plugin(i_ctx_p, pbfont->FAPI_ID, &I));
    scale = 1 << I->frac_shift;
    unscale = 1 << -path_interface.shift;
    /* Note: ctm here is upside down. */
    matrix[0] =  (FracInt)(ctm->xx / base_font_matrix->xx * 72 / dev->HWResolution[0] * scale);
    matrix[1] = -(FracInt)(ctm->xy / base_font_matrix->yy * 72 / dev->HWResolution[0] * scale);
    matrix[2] =  (FracInt)(ctm->yx / base_font_matrix->xx * 72 / dev->HWResolution[1] * scale);
    matrix[3] = -(FracInt)(ctm->yy / base_font_matrix->yy * 72 / dev->HWResolution[1] * scale);
    matrix[4] =  (FracInt)(ctm->tx / base_font_matrix->xx * 72 / dev->HWResolution[0] * scale);
    matrix[5] =  (FracInt)(ctm->ty / base_font_matrix->yy * 72 / dev->HWResolution[1] * scale);
    HWResolution[0] = (FracInt)(dev->HWResolution[0] * scale);
    HWResolution[1] = (FracInt)(dev->HWResolution[1] * scale);
    path_interface.shift = _fixed_shift - I->frac_shift;
    path_interface.olh->path = penum_s->pgs->show_gstate->path;
    path_interface.olh->x0 = penum_s->pgs->ctm.tx_fixed;
    path_interface.olh->y0 = penum_s->pgs->ctm.ty_fixed;
    if (dict_find_string(osp - 1, "SubfontId", &SubfontId) >= 0 && r_has_type(SubfontId, t_integer))
        subfont = SubfontId->value.intval;
    CheckRET(renderer_retcode(i_ctx_p, I, (I->get_font)(&I->ig, font_file_path, subfont, matrix, HWResolution, xlatmap, bCacheFonts, &API_font_handle)));
    if (char_by_name) {
        bool can_retrive_char_by_name = false;
        CheckRET(renderer_retcode(i_ctx_p, I, (I->can_retrive_char_by_name)(&I->ig, &API_font_handle, &can_retrive_char_by_name)));
        obj_string_data(&char_name, &cr.char_name, &cr.char_name_length);
        if (!can_retrive_char_by_name) {
	    /* Translate from char name to 3d party encoding : */
	    ref *Decoding, *char_code;
            if (dict_find_string(osp - 1, "Decoding", &Decoding) >= 0 && r_has_type(Decoding, t_dictionary)) {
	        if (dict_find(Decoding, &char_name, &char_code) >= 0 && r_has_type(char_code, t_integer))
		    int_param(char_code, 0xFFFF, &cr.char_code);
	    }
        }
    }
    if (SHOW_IS(penum, TEXT_DO_NONE)) {
	CheckRET(renderer_retcode(i_ctx_p, I, (I->get_char_width)(&I->ig, API_font_handle, &cr, &metrics.esc_x, &metrics.esc_y)));
    } else if (igs->in_charpath) {
	CheckRET(renderer_retcode(i_ctx_p, I, (I->outline_char)(&I->ig, API_font_handle, &cr, &path_interface, &metrics)));
    } else if (I->renderer_caches_chars) {
	CheckRET(renderer_retcode(i_ctx_p, I, (I->get_char_raster_metrics)(&I->ig, API_font_handle, &cr, &metrics)));
	{   gx_device_memory * mdev;
	    FAPI_raster rast;
	    fixed dx = 0; /* fixme : what is it ? */
	    const gx_drawing_color * pdcolor = penum->pdcolor;
	    const gx_clip_path * pcpath = i_ctx_p->pgs->clip_path; 
	    double orig_x = metrics.orig_x / 16.0;
	    double orig_y = metrics.orig_y / 16.0;
	    CheckRET(renderer_retcode(i_ctx_p, I, (I->get_char_raster)(&I->ig, API_font_handle, &cr, &rast)));
	    if (metrics.bbox_x1 >= metrics.bbox_x0)
		CheckRET(dev_proc(dev, fill_mask)(dev, rast.p, dx, rast.line_step, 0,
				  (int)(penum_s->pgs->ctm.tx + orig_x + 0.5), 
				  (int)(penum_s->pgs->ctm.ty - orig_y + 0.5), 
				  rast.width, rast.height,
				  pdcolor, 1, rop3_default, pcpath));
	}
	CheckRET(renderer_retcode(i_ctx_p, I, (I->release_char_raster)(&I->ig, API_font_handle, &cr)));
    } else
	return_error(e_unregistered); /* unimplemented */
    penum_s->wxy.x =   import_shift(metrics.esc_x, path_interface.shift);
    penum_s->wxy.y = - import_shift(metrics.esc_y, path_interface.shift);
    penum_s->width_status = sws_no_cache;
    pop(2);
    return 0;
}

private int FAPI_char(i_ctx_t *i_ctx_p)
{   /* Stack : <font> <code|name> --> - */
    ref *v;
    char *font_file_path;
    int code;
    gx_device *dev = gs_currentdevice_inline(igs);
    if (dict_find_string(osp - 1, "Path", &v) < 0 || !r_has_type(v, t_string))
	return_error(e_invalidfont);
    font_file_path = ref_to_string(v, imemory, "font file path");
    code = FAPI_do_char(i_ctx_p, dev, font_file_path);
    gs_free_string(imemory, (byte *)font_file_path, r_size(v) + 1, "font file path");
    return code;
}

/* <font> <code> .FAPIBuildChar - */
private int zFAPIBuildChar(i_ctx_t *i_ctx_p)
{   check_type(*osp, t_integer);
    return FAPI_char(i_ctx_p);
}

/* <font> <name> .FAPIBuildGlyph - */
private int zFAPIBuildGlyph(i_ctx_t *i_ctx_p)
{   check_type(*osp, t_name);
    return FAPI_char(i_ctx_p);
}

private int do_passfonttoFAPI(i_ctx_t *i_ctx_p, char *font_file_path, bool *success)
{   ref *pdr = osp;  /* font dict */
    dict *pdict = pdr->value.pdict;
    int HWResolution[2], BBox[4];
    FracInt matrix[6] = {1, 0, 0, 1, 0, 0}; /* to be scaled */
    FAPI_metrics metrics;
    void *API_font_handle;
    gs_plugin_interface **plugins = i_get_installed_plugins(i_ctx_p);
    int i, subfont = 0;
    char *xlatmap;
    ref *SubfontId;
    CheckRET(FAPI_get_xlatmap(i_ctx_p, &xlatmap));
    if (dict_find_string(pdr, "SubfontId", &SubfontId) >= 0 && r_has_type(SubfontId, t_integer))
        subfont = SubfontId->value.intval;
    *success = false;
    for (i = 0; plugins[i] != 0; i++) {
	ref FAPI_ID;
        FAPI_interface *I;
       	if (strcmp(plugins[i]->factory->type, "FAPI"))
            continue;
        I = (FAPI_interface *)plugins[i];
        if (I->ig.factory->version_major < 1)
            return renderer_retcode(i_ctx_p, I, (int)gs_plugin__rc_version);
        CheckRET(renderer_retcode(i_ctx_p, I, I->ensure_open(&I->ig)));
	HWResolution[0] = HWResolution[1] = 72 << I->frac_shift;
	matrix[0] = matrix[3] = 1 << I->frac_shift;
	if((I->get_font)(&I->ig, font_file_path, subfont, matrix, HWResolution, xlatmap, bCacheFonts, &API_font_handle))
	  continue;
	if((I->get_font_bbox)(&I->ig, &API_font_handle, BBox)) /* Try to do something to check whether the font is valid. */
	  continue;
	/* We found a good renderer, so go with it : */
        CheckRET(name_ref((unsigned char *)I->ig.factory->subtype, strlen(I->ig.factory->subtype), &FAPI_ID, false));
        /* Insert entries to font dictionary : */
	CheckRET(dict_put_string(pdr, "FAPI", &FAPI_ID, NULL)); 
	CheckRET(dict_put_string(pdr, "Path", osp - 1, NULL)); /* Path to the font */
	CheckRET(dict_put_string(pdr, "Temp", osp - 1, NULL)); /* Mark it temporary */
	*success = true;
	return 0;
    }
    /* Could not find renderer, return with success=false. */
    return 0;
}

/* <diskpath> <font_dict> .passfonttoFAPI bool */
private int zpassfonttoFAPI(i_ctx_t *i_ctx_p)
{   /* must insert /Path and /FAPI to font dictionary */
    os_ptr op = osp;
    char *font_file_path;
    int code;
    bool found;
    ref rfound;
    check_type(*(op - 1), t_string);
    check_type(*op, t_dictionary);
    font_file_path = ref_to_string(osp - 1, imemory, "font file path");
    code = do_passfonttoFAPI(i_ctx_p, font_file_path, &found);
    gs_free_string(imemory, (byte *)font_file_path, r_size(op - 1) + 1, "font file path");
    if(code != 0)
	return code;
    pop(2);
    op = osp;
    push(1);
    make_bool(op, found);
    return 0;
}

/*  The following operators are subject of further replacement with
    totally different code. Users must not call them.
*/

/* <dicct1> <dict2> ..copydict <dict2>
/* copy for dictionaries, ignoring read attribute. */
private int zzcopydict(i_ctx_t *i_ctx_p)
{   os_ptr op = osp;
    os_ptr op1 = op - 1;
    check_type(*op,  t_dictionary);
    check_type(*op1, t_dictionary);
    check_dict_write(*op);
    CheckRET(idict_copy(op1, op));
    ref_assign(op1, op);
    pop(1);
    return 0;
}

/* <obj> .cvrd <obj> */
/* Allow read access to obj */
private int zzard(i_ctx_t *i_ctx_p)
{   os_ptr op = osp;
    r_set_attrs(op, a_read);
    return 0;
}

/* <array|dict|name|packedarray|string> ..length <int> */
/* Length of object, ignoring read attribute. */
private int zzlength(i_ctx_t *i_ctx_p)
{   os_ptr op = osp;
    switch (r_type(op)) {
	case t_array:
	case t_string:
	case t_mixedarray:
	case t_shortarray:
	    make_int(op, r_size(op));
	    return 0;
	case t_dictionary:
	    make_int(op, dict_length(op));
	    return 0;
	case t_name: {
	    ref str;
	    name_string_ref(op, &str);
	    make_int(op, r_size(&str));
	    return 0;
	}
	case t_astruct:
	    if (gs_object_type(imemory, op->value.pstruct) != &st_bytes)
		return_error(e_typecheck);
	    make_int(op, gs_object_size(imemory, op->value.pstruct));
	    return 0;
	default:
	    return_op_typecheck(op);
    }
}

/* prefix .tempfile <filename> <file> */
private int ztempfile(i_ctx_t *i_ctx_p)
{   os_ptr op = osp;
    byte *filename1, *filename2, *buffer;
    char fname[gp_file_name_sizeof]; /* Fixme : not sure that this size is sufficient for Windows. */
    int l;
    stream *s;
    FILE *file;
    if (r_type(op) != t_string)
        return_op_typecheck(op);
    {   byte * prefix = (byte *)ref_to_string(op, imemory, "file name prefix");
        if (prefix == 0)
	    return_error(e_VMerror);
        file = gp_open_scratch_file((char *)prefix, fname, "wb");
        gs_free_string(imemory, prefix, r_size(op) + 1, "file name prefix");
    }
    if (file==NULL)
        return_error(e_undefinedfilename);
    l = strlen(fname);
    filename1 = ialloc_string(l, "temp file name");
    if (filename1 == 0)
        goto e1;
    make_string(op, a_read | icurrent_space, l, filename1);
    filename2 = ialloc_string(l + 1, "temp file name");
    if (filename2 == 0)
        goto e1;
    memcpy(filename1, fname, l);
    strcpy(filename2, fname);
    buffer = gs_alloc_bytes(imemory, file_default_buffer_size, "ztempfile(buffer)");
    if (buffer == 0)
        goto e2;
    s = file_alloc_stream(imemory, "file_open_stream");
    if (s == 0)
        goto e3;
    swrite_file(s, file, buffer, file_default_buffer_size);
    s->file_modes |= s_mode_write;
    s->file_name.data = filename2;
    s->file_name.size = l + 1;
    push(1);
    make_stream_file(op, s, "wb");
    return 0;

e3: gs_free_object(imemory, buffer, "ztempfile(buffer)");
e2: ifree_string(filename1, l, "temp file name");
e1: fclose(file);
    return_error(e_VMerror);
}

const op_def zfontAPI_op_defs[] =
{   {"2.buildfontFAPI", zbuildfontFAPI},
    {"2.FAPIBuildChar", zFAPIBuildChar},
    {"2.FAPIBuildGlyph", zFAPIBuildGlyph},
    {"2.passfonttoFAPI", zpassfonttoFAPI},
    {"2..copydict", zzcopydict},
    {"2..cvrd", zzard},
    {"2..length", zzlength},
    {"2.tempfile", ztempfile},
    op_def_end(0)
};

